// ----------------------------------------------------------------------------
// Input.java
// Line-oriented input from standard input stream.
//
// Author:  Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Created: August 19, 2001; Last modified: June 7, 2013
// ----------------------------------------------------------------------------
package kwm;

import java.io.*;

public class Input
{
  // last read has returned end of stream
  private static boolean endFlag = false;

  // message describing the error
  private static String errorMessage = null;

  // buffered reader wrapped around standard input stream
  private static BufferedReader stdReader;

  // current reader
  private static BufferedReader reader;

  // current encoding and standard encoding
  private static String stdEncoding = "Cp1252";

  // set encoding to standard encoding for MS Windows
  static
  {
    setEncoding(stdEncoding);
    stdReader = reader;
  }

  // -------------------------------------------------------------------------
  // setEncoding(encoding)
  // set input encoding
  //
  // 'encoding' is the name of an encoding implemented on the machine
  // - 'Cp1252' for MS Windows (default)
  // - 'UTF8' for Linux and Mac OS X
  // -------------------------------------------------------------------------
  public static void setEncoding(String encoding)
  {
    try
    {
      Reader r = new InputStreamReader(System.in, encoding);
      reader = new BufferedReader(r);
    }
    catch(UnsupportedEncodingException e)
    {
      System.out.println("Unsupported encoding:");
      System.out.println(e);
      System.out.println("Encoding remains unchanged.");
    }
  }

  // -------------------------------------------------------------------------
  // test program: iteratively
  // - reads character [B|b|s|i|l|f|d|S]
  // - reads element of type denoted by character
  // - prints element respectively error message
  // terminating by pressing <ctrl-d> (end of input)
  // -------------------------------------------------------------------------
  public static void main(String[] args)
  {
    if (args.length == 1)
    {
      openInput(args[0]);
      if (!Input.isOkay())
      {
        System.out.println("File '" + args[0] + "' could not be opened.");
        System.out.println("Error: " + Input.getError());
        return;
      }
    }
    while (true)
    {
      System.out.print("Enter [B|c|b|s|i|l|f|d|S]: ");
      char c = readChar();
      if (hasEnded()) break;
      if (!isOkay())
      {
        clearError();
        continue;
      }
      switch (c)
      {
      case 'B':
      {
        System.out.print("Enter boolean: ");
        boolean b = readBoolean();
        if (isOkay()) System.out.println("Okay: " + b);
        break;
      }
      case 'c':
      {
        System.out.print("Enter char: ");
        char ch = readChar();
        if (isOkay()) System.out.println("Okay: " + ch);
        break;
      }
      case 'b':
      {
        System.out.print("Enter byte: ");
        byte b = readByte();
        if (isOkay()) System.out.println("Okay: " + b);
        break;
      }
      case 's':
      {
        System.out.print("Enter short: ");
        short s = readShort();
        if (isOkay()) System.out.println("Okay: " + s);
         break;
      }
      case 'i':
      {
        System.out.print("Enter int: ");
        int i = readInt();
        if (isOkay()) System.out.println("Okay: " + i);
        break;
      }
      case 'l':
      {
        System.out.print("Enter long: ");
        long l = readLong();
        if (isOkay()) System.out.println("Okay: " + l);
        break;
      }
      case 'f':
      {
        System.out.print("Enter float: ");
        float f = readFloat();
        if (isOkay()) System.out.println("Okay: " + f);
        break;
      }
      case 'd':
      {
        System.out.print("Enter double: ");
        double d = readDouble();
        if (isOkay()) System.out.println("Okay: " + d);
        break;
      }
      case 'S':
      {
        System.out.print("Enter string: ");
        String s = readString();
        if (isOkay()) System.out.println("Okay: " + s);
        break;
      }
      }
      if (!isOkay())
      {
        if (hasEnded()) return;
        System.out.println("Error: " + getError());
        clearError();
      }
    }
    if (args.length == 1)
    {
      closeInput();
      if (!Input.isOkay())
      {
        System.out.println("File could not be closed.");
        System.out.println("Error: " + Input.getError());
        return;
      }
    }
    System.out.println("\nTerminated.");
  }

  // -------------------------------------------------------------------------
  // openInput(name);
  // open a file "name" for input
  //
  // Input condition:
  //   "name" is not null.
  // Effect:
  //   if "isOkay()", then a file called "name" exists and could be opened
  //   for reading; all further read operations refer to the denoted file
  //   rather than to the standard input.
  //   if "!isOkay()", all read operations still refer to the standard input.
  // Note:
  //   "name" may be a complete file path (including directories).
  // -------------------------------------------------------------------------
  public static void openInput(String name)
  {
    try
    {
      reader = new BufferedReader(new FileReader(name));
    }
    catch(FileNotFoundException e)
    {
      errorMessage = e.toString();
    }
  }

  // -------------------------------------------------------------------------
  // closeInput();
  // close the current input file.
  //
  // Effect:
  //   "isOkay()" iff the current input has referred to a file which could
  //   be successfully closed.
  //   in any case, all further reads refer to the standard input stream.
  // -------------------------------------------------------------------------
  public static void closeInput()
  {
    if (reader == stdReader)
    {
      errorMessage = "No file has been opened!";
      return;
    }
    try
    {
      reader.close();
    }
    catch (IOException e)
    {
      errorMessage = e.toString();
    }
    reader = stdReader;
    endFlag = false;
  }

  // -------------------------------------------------------------------------
  // b = isOkay()
  // indicates whether last read was successful.
  //
  // Output condition:
  //   "b" is true iff last read returned a meaningful value.
  //   if "b" is false, then either "hasEnded()" is true or "getError()"
  //     contains a message describing the error.
  // -------------------------------------------------------------------------
  public static boolean isOkay()
  {
    return !endFlag && errorMessage == null;
  }

  // -------------------------------------------------------------------------
  // b = hasEnded()
  // indicates whether in last read end of stream occurred.
  //
  // Output condition:
  //   "b" is true iff in last read end of stream occurred.
  // -------------------------------------------------------------------------
  public static boolean hasEnded()
  {
    return endFlag;
  }

  // -------------------------------------------------------------------------
  // s = getError()
  // indicates error occuring in last read.
  //
  // Input condition:
  //   "isOkay()" is false and "hasEnded()" is false.
  // Output condition:
  //   "s" is string describing message.
  // -------------------------------------------------------------------------
  public static String getError()
  {
    return errorMessage;
  }

  // -------------------------------------------------------------------------
  // clearError()
  // clears error that occurred in last read.
  //
  // Input condition
  //   "isOkay()" is false and "hasEnded()" is false.
  // Output condition:
  //   "isOkay()" is true.
  // -------------------------------------------------------------------------
  public static void clearError()
  {
    errorMessage = null;
  }

  // -------------------------------------------------------------------------
  // l = readLine()
  // reads line terminated by end of line character from standard input.
  //
  // Input condition:
  //   "errorMessage" is null.
  // Output condition:
  //   if reading was okay, "l" is line read without end of line character.
  //   if end of input stream or read error occurred, "l" is null.
  //   if end of input stream occurred, "endFlag" is true
  //   if read error occurred, "errorMessage" is a string
  //     describing the reason of the error.
  // -------------------------------------------------------------------------
  private static String readLine()
  {
    try
    {
      String line = reader.readLine();
      if (line == null)
  endFlag = true;
      return line;
    }
    catch(Exception e)
    {
      errorMessage = e.toString();
    }
    return null;
  }

  // -------------------------------------------------------------------------
  // b = readBoolean()
  // reads line with boolean value from standard input.
  // value is denoted by line whose content is "true" or "false".
  //
  // Input condition:
  //   "hasEnded()" is false.
  // Output condition:
  //   if "isOkay()", then "b" is a boolean value read from standard input.
  // -------------------------------------------------------------------------
  public static boolean readBoolean()
  {
    String line = readLine();
    if (line == null) return false;
    if (line.equals("true")) return true;
    if (line.equals("false")) return false;
    errorMessage = "Invalid boolean string \"" + line + "\"";
    return false;
  }

  // -------------------------------------------------------------------------
  // c = readChar()
  // reads line with character value from standard input.
  // value is denoted by line whose content is a single character.
  //
  // Input condition:
  //   "hasEnded()" is false.
  // Output condition:
  //   if "isOkay()", then "c" is a character value read from standard input.
  // -------------------------------------------------------------------------
  public static char readChar()
  {
    String line = readLine();
    if (line == null) return 0;
    if (line.length() == 1) return line.charAt(0);
    errorMessage = "Invalid character string \"" + line + "\"";
    return 0;
  }

  // -------------------------------------------------------------------------
  // b = readByte()
  // reads line with byte value from standard input.
  // value is denoted by line whose content is an integer number
  // in decimal representation.
  //
  // Input condition:
  //   "hasEnded()" is false.
  // Output condition:
  //   if "isOkay()", then "b" is a byte value read from standard input.
  // -------------------------------------------------------------------------
  public static byte readByte()
  {
    String line = readLine();
    if (line == null) return Byte.MIN_VALUE;
    try
    {
      int i =  Integer.parseInt(line);
      if (Byte.MIN_VALUE <= i && i <= Byte.MAX_VALUE) return (byte)i;
      errorMessage = "Outside byte range: " + line;
    }
    catch(Exception e)
    {
      errorMessage = e.toString();
    }
    return Byte.MIN_VALUE;
  }

  // -------------------------------------------------------------------------
  // s = readShort()
  // reads line with short value from standard input.
  // value is denoted by line whose content is an integer number
  // in decimal representation.
  //
  // Input condition:
  //   "hasEnded()" is false.
  // Output condition:
  //   if "isOkay()", then "s" is a short value read from standard input.
  // -------------------------------------------------------------------------
  public static short readShort()
  {
    String line = readLine();
    if (line == null) return Short.MIN_VALUE;
    try
    {
      int i =  Integer.parseInt(line);
      if (Short.MIN_VALUE <= i && i <= Short.MAX_VALUE) return (short)i;
      errorMessage = "Outside short range: " + line;
    }
    catch(Exception e)
    {
      errorMessage = e.toString();
    }
    return Short.MIN_VALUE;
  }

  // -------------------------------------------------------------------------
  // i = readInt()
  // reads line with integer value from standard input.
  // value is denoted by line whose content is an integer number
  // in decimal representation.
  //
  // Input condition:
  //   "hasEnded()" is false.
  // Output condition:
  //   if "isOkay()", then "i" is an integer value read from standard input.
  // -------------------------------------------------------------------------
  public static int readInt()
  {
    String line = readLine();
    if (line == null) return Integer.MIN_VALUE;
    try
    {
      return Integer.parseInt(line);
    }
    catch(Exception e)
    {
      errorMessage = e.toString();
    }
    return Integer.MIN_VALUE;
  }

  // -------------------------------------------------------------------------
  // l = readLong()
  // reads line with long value from standard input.
  // value is denoted by line whose content is an integer number
  // in decimal representation.
  //
  // Input condition:
  //   "hasEnded()" is false.
  // Output condition:
  //   if "isOkay()", then "l" is a long value read from standard input.
  // -------------------------------------------------------------------------
  public static long readLong()
  {
    String line = readLine();
    if (line == null) return Long.MIN_VALUE;
    try
    {
      return Long.parseLong(line);
    }
    catch(Exception e)
    {
      errorMessage = e.toString();
    }
    return Long.MIN_VALUE;
  }

  // -------------------------------------------------------------------------
  // f = readFloat()
  // reads line with float value from standard input.
  // value is denoted by line whose content is an float number
  // in decimal representation.
  //
  // Input condition:
  //   "hasEnded()" is false.
  // Output condition:
  //   if "isOkay()", then "f" is a float value read from standard input.
  // -------------------------------------------------------------------------
  public static float readFloat()
  {
    String line = readLine();
    if (line == null) return Float.MIN_VALUE;
    try
    {
      // do not use Float.parseFloat() because of J++ compatibility
      return Float.valueOf(line).floatValue();
    }
    catch(Exception e)
    {
      errorMessage = e.toString();
    }
    return Float.MIN_VALUE;
  }

  // -------------------------------------------------------------------------
  // d = readFloat()
  // reads line with float value from standard input.
  // value is denoted by line whose content is an float number
  // in decimal representation.
  //
  // Input condition:
  //   "hasEnded()" is false.
  // Output condition:
  //   if "isOkay()", then "d" is a float value read from standard input.
  // -------------------------------------------------------------------------
  public static double readDouble()
  {
    String line = readLine();
    if (line == null) return Double.MIN_VALUE;
    try
    {
      // do not use Double.parseDouble() because of J++ compatibility
      return Double.valueOf(line).doubleValue();
    }
    catch(Exception e)
    {
      errorMessage = e.toString();
    }
    return Double.MIN_VALUE;
  }

  // -------------------------------------------------------------------------
  // s = readString()
  // reads line from standard input.
  //
  // Input condition:
  //   "hasEnded()" is false.
  // Output condition:
  //   if "isOkay()", then "s" is a line read from standard input
  //     (without line termination character).
  // -------------------------------------------------------------------------
  public static String readString()
  {
    return readLine();
  }
}
